@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2020 -- leonerd@leonerd.org.uk
 
-package Object::Pad::MOP::Method 0.78;
+package Object::Pad::MOP::Method 0.808;
 
 use v5.14;
 use warnings;
@@ -31,18 +31,20 @@ effect. They can be silenced with
 
 =head2 name
 
-   $name = $metamethod->name
+   $name = $metamethod->name;
 
 Returns the name of the method, as a plain string.
 
 =head2 class
 
+   $metaclass = $metamethod->class;
+
 Returns the L<Object::Pad::MOP::Class> instance representing the class of
 which this method is a member.
 
 =head2 is_common
 
-   $bool = $metamethod->is_common
+   $bool = $metamethod->is_common;
 
 I<Since version 0.62.>
 
