@@ -5,12 +5,6 @@
 #include "perl.h"
 #include "XSUB.h"
 
-#include "object_pad.h"
-#include "class.h"
-#include "field.h"
-
-#undef register_field_attribute
-
 #include "perl-backcompat.c.inc"
 #include "perl-additions.c.inc"
 #include "force_list_keeping_pushmark.c.inc"
@@ -19,6 +13,12 @@
 #include "newOP_CUSTOM.c.inc"
 #include "OP_HELEMEXISTSOR.c.inc"
 
+#include "object_pad.h"
+#include "class.h"
+#include "field.h"
+
+#undef register_field_attribute
+
 #if HAVE_PERL_VERSION(5,36,0)
 #  define HAVE_OP_WEAKEN
 #endif
@@ -26,17 +26,18 @@
 #define need_PLparser()  ObjectPad__need_PLparser(aTHX)
 void ObjectPad__need_PLparser(pTHX); /* in Object/Pad.xs */
 
-FieldMeta *ObjectPad_mop_create_field(pTHX_ SV *fieldname, ClassMeta *classmeta)
+FieldMeta *ObjectPad_mop_create_field(pTHX_ SV *fieldname, FIELDOFFSET fieldix, ClassMeta *classmeta)
 {
   FieldMeta *fieldmeta;
   Newx(fieldmeta, 1, FieldMeta);
 
-  assert(classmeta->next_fieldix > -1);
+  assert(fieldix > -1);
 
   *fieldmeta = (FieldMeta){
-    .name    = SvREFCNT_inc(fieldname),
-    .class   = classmeta,
-    .fieldix = classmeta->next_fieldix,
+    .name      = SvREFCNT_inc(fieldname),
+    .is_direct = true,
+    .class     = classmeta,
+    .fieldix   = fieldix,
   };
 
   return fieldmeta;
@@ -221,14 +222,11 @@ void ObjectPad_mop_field_apply_attribute(pTHX_ FieldMeta *fieldmeta, const char
   av_push(fieldmeta->hooks, (SV *)hook);
 }
 
-struct FieldHook *ObjectPad_mop_field_get_attribute(pTHX_ FieldMeta *fieldmeta, const char *name)
+static FieldAttributeRegistration *get_active_registration(pTHX_ const char *name)
 {
   COPHH *cophh = CopHINTHASH_get(PL_curcop);
 
-  /* First, work out what hookfuncs the name maps to */
-
-  FieldAttributeRegistration *reg;
-  for(reg = fieldattrs; reg; reg = reg->next) {
+  for(FieldAttributeRegistration *reg = fieldattrs; reg; reg = reg->next) {
     if(!strEQ(name, reg->name))
       continue;
 
@@ -236,9 +234,17 @@ struct FieldHook *ObjectPad_mop_field_get_attribute(pTHX_ FieldMeta *fieldmeta,
         !cophh_fetch_pvn(cophh, reg->funcs->permit_hintkey, reg->permit_hintkeylen, 0, 0))
       continue;
 
-    break;
+    return reg;
   }
 
+  return NULL;
+}
+
+struct FieldHook *ObjectPad_mop_field_get_attribute(pTHX_ FieldMeta *fieldmeta, const char *name)
+{
+  /* First, work out what hookfuncs the name maps to */
+  FieldAttributeRegistration *reg = get_active_registration(aTHX_ name);
+
   if(!reg)
     return NULL;
 
@@ -260,21 +266,8 @@ struct FieldHook *ObjectPad_mop_field_get_attribute(pTHX_ FieldMeta *fieldmeta,
 
 AV *ObjectPad_mop_field_get_attribute_values(pTHX_ FieldMeta *fieldmeta, const char *name)
 {
-  COPHH *cophh = CopHINTHASH_get(PL_curcop);
-
   /* First, work out what hookfuncs the name maps to */
-
-  FieldAttributeRegistration *reg;
-  for(reg = fieldattrs; reg; reg = reg->next) {
-    if(!strEQ(name, reg->name))
-      continue;
-
-    if(reg->funcs->permit_hintkey &&
-        !cophh_fetch_pvn(cophh, reg->funcs->permit_hintkey, reg->permit_hintkeylen, 0, 0))
-      continue;
-
-    break;
-  }
+  FieldAttributeRegistration *reg = get_active_registration(aTHX_ name);
 
   if(!reg)
     return NULL;
@@ -302,10 +295,84 @@ AV *ObjectPad_mop_field_get_attribute_values(pTHX_ FieldMeta *fieldmeta, const c
   return ret;
 }
 
+SV *ObjectPad_get_obj_fieldsv(pTHX_ SV *self, ClassMeta *classmeta, FieldMeta *fieldmeta)
+{
+  SV *fieldstore;
+  FIELDOFFSET fieldix;
+
+  assert(SvROK(self));
+  assert(SvOBJECT(SvRV(self)));
+
+  if(classmeta->type == METATYPE_ROLE) {
+    HV *objstash = SvSTASH(SvRV(self));
+    const char *key = HvNAME(objstash);
+    STRLEN klen = HvNAMELEN(objstash);
+    if(HvNAMEUTF8(objstash))
+      klen = -klen;
+
+    assert(key);
+    SV **svp = hv_fetch(classmeta->role.applied_classes, key, klen, 0);
+    if(!svp)
+      croak("Cannot fetch role field value from a non-applied instance");
+
+    RoleEmbedding *embedding = (RoleEmbedding *)*svp;
+
+    fieldstore = get_obj_fieldstore(self, embedding->classmeta->repr, true);
+    fieldix = fieldmeta->fieldix + embedding->offset;
+  }
+  else {
+    const char *stashname = HvNAME(classmeta->stash);
+
+    if(!stashname || !sv_derived_from(self, stashname))
+      croak("Cannot fetch field value from a non-derived instance");
+
+    fieldstore = get_obj_fieldstore(self, classmeta->repr, true);
+    fieldix = fieldmeta->fieldix;
+  }
+
+  if(fieldix > fieldstore_maxfield(fieldstore))
+    croak("ARGH: instance does not have a field at index %ld", (long int)fieldix);
+
+  SV *sv = fieldstore_fields(fieldstore)[fieldix];
+
+  return sv;
+}
+
+static OP *pp_fieldsv(pTHX)
+{
+  dSP;
+  FIELDOFFSET fieldix = PL_op->op_targ;
+  if(PL_op->op_flags & OPf_SPECIAL) {
+    RoleEmbedding *embedding = get_embedding_from_pad();
+
+    if(embedding && embedding != &ObjectPad__embedding_standalone) {
+      fieldix += embedding->offset;
+    }
+  }
+
+  SV *fieldstore = PAD_SVl(PADIX_FIELDS);
+
+  SV *fieldsv = fieldstore_fields(fieldstore)[fieldix];
+
+  EXTEND(SP, 1);
+  PUSHs(fieldsv);
+
+  RETURN;
+}
+
+#define newFIELDSVOP(flags, fieldix)  S_newFIELDSVOP(aTHX_ flags, fieldix)
+static OP *S_newFIELDSVOP(pTHX_ U32 flags, FIELDOFFSET fieldix)
+{
+  OP *o = newOP_CUSTOM(&pp_fieldsv, flags);
+  o->op_targ = fieldix;
+  return o;
+}
+
 #define gen_field_init_op(fieldmeta)  S_gen_field_init_op(aTHX_ fieldmeta)
 static OP *S_gen_field_init_op(pTHX_ FieldMeta *fieldmeta)
 {
   ClassMeta *classmeta = fieldmeta->class;
+  U8 opf_special_if_role = (classmeta->type == METATYPE_ROLE) ? OPf_SPECIAL : 0;
 
   char sigil = SvPV_nolen(fieldmeta->name)[0];
   OP *op = NULL;
@@ -348,9 +415,7 @@ static OP *S_gen_field_init_op(pTHX_ FieldMeta *fieldmeta)
         op = newBINOP(OP_SASSIGN, 0,
           valueop,
           /* $fields[$idx] */
-          newAELEMOP(OPf_MOD,
-            newPADxVOP(OP_PADAV, OPf_MOD|OPf_REF, PADIX_SLOTS),
-            fieldmeta->fieldix));
+          newFIELDSVOP(OPf_MOD | opf_special_if_role, fieldmeta->fieldix));
       break;
     }
     case '@':
@@ -366,9 +431,7 @@ static OP *S_gen_field_init_op(pTHX_ FieldMeta *fieldmeta)
       if(valueop) {
         /* $fields[$idx]->@* or ->%* */
         OP *lhs = force_list_keeping_pushmark(newUNOP(coerceop, OPf_MOD|OPf_REF,
-                    newAELEMOP(0,
-                      newPADxVOP(OP_PADAV, OPf_MOD|OPf_REF, PADIX_SLOTS),
-                      fieldmeta->fieldix)));
+                    newFIELDSVOP(opf_special_if_role, fieldmeta->fieldix)));
 
         op = newBINOP(OP_AASSIGN, 0,
             force_list_keeping_pushmark(valueop),
@@ -504,6 +567,7 @@ static void S_generate_field_accessor_method(pTHX_ FieldMeta *fieldmeta, SV *mna
   ENTER;
 
   ClassMeta *classmeta = fieldmeta->class;
+  U8 opf_special_if_role = (classmeta->type == METATYPE_ROLE ? OPf_SPECIAL : 0);
   char sigil = SvPVX(fieldmeta->name)[0];
 
   SV *mname_fq = newSVpvf("%" SVf "::%" SVf, classmeta->name, mname);
@@ -523,16 +587,15 @@ static void S_generate_field_accessor_method(pTHX_ FieldMeta *fieldmeta, SV *mna
 
   extend_pad_vars(classmeta);
 
-  struct AccessorGenerationCtx ctx = { 0 };
-
-  ctx.padix = pad_add_name_sv(fieldmeta->name, 0, NULL, NULL);
+  PADOFFSET padix = pad_add_name_sv(fieldmeta->name, 0, NULL, NULL);
   intro_my();
 
   OP *ops = op_append_list(OP_LINESEQ, NULL,
     newSTATEOP(0, NULL, NULL));
+  OP *methstartop;
   ops = op_append_list(OP_LINESEQ, ops,
-    newMETHSTARTOP(0 |
-      (classmeta->type == METATYPE_ROLE ? OPf_SPECIAL : 0) |
+    methstartop = newMETHSTARTOP(0 |
+      opf_special_if_role |
       (classmeta->repr << 8)));
 
   int req_args = 0;
@@ -554,16 +617,122 @@ static void S_generate_field_accessor_method(pTHX_ FieldMeta *fieldmeta, SV *mna
   ops = op_append_list(OP_LINESEQ, ops,
     make_argcheck_ops(req_args, opt_args, slurpy_arg, mname_fq));
 
-  U32 flags = 0;
+  FIELDOFFSET fieldix = fieldmeta->fieldix;
+
+  U8 private = 0;
 
   switch(sigil) {
-    case '$': flags = OPpFIELDPAD_SV << 8; break;
-    case '@': flags = OPpFIELDPAD_AV << 8; break;
-    case '%': flags = OPpFIELDPAD_HV << 8; break;
+    case '$': private = OPpFIELDPAD_SV; break;
+    case '@': private = OPpFIELDPAD_AV; break;
+    case '%': private = OPpFIELDPAD_HV; break;
   }
 
-  ops = op_append_list(OP_LINESEQ, ops,
-    newFIELDPADOP(flags, ctx.padix, fieldmeta->fieldix));
+#ifdef METHSTART_CONTAINS_FIELD_BINDINGS
+  {
+    UNOP_AUX_item *aux;
+    Newx(aux, 2 + 1*2, UNOP_AUX_item);
+    cUNOP_AUXx(methstartop)->op_aux = aux;
+
+    (aux++)->uv = 1;       /* fieldcount */
+    (aux++)->uv = fieldix; /* max_fieldix */
+
+    (aux++)->uv = padix;
+    (aux++)->uv = ((UV)private << FIELDIX_TYPE_SHIFT) | fieldix;
+  }
+#else
+  {
+    ops = op_append_list(OP_LINESEQ, ops,
+      newFIELDPADOP(private << 8 | opf_special_if_role, padix, fieldix));
+  }
+#endif
+
+  /* Generate the basic ops here so the ordering doesn't matter if other
+   * attributes want to modify these */
+
+  struct AccessorGenerationCtx ctx = {
+    .padix = padix,
+  };
+
+  switch(type) {
+    case ACCESSOR_READER:
+    {
+      OPCODE optype = 0;
+
+      switch(sigil) {
+        case '$': optype = OP_PADSV; break;
+        case '@': optype = OP_PADAV; break;
+        case '%': optype = OP_PADHV; break;
+      }
+
+      ctx.retop = newLISTOP(OP_RETURN, 0,
+        newOP(OP_PUSHMARK, 0),
+        newPADxVOP(optype, 0, padix));
+
+      break;
+    }
+
+    case ACCESSOR_WRITER:
+    {
+      switch(sigil) {
+        case '$':
+          ctx.bodyop = newBINOP(OP_SASSIGN, 0,
+            newOP(OP_SHIFT, 0),
+            newPADxVOP(OP_PADSV, 0, padix));
+          break;
+
+        case '@':
+          ctx.bodyop = newBINOP(OP_AASSIGN, 0,
+            force_list_keeping_pushmark(newUNOP(OP_RV2AV, 0, newGVOP(OP_GV, 0, PL_defgv))),
+            force_list_keeping_pushmark(newPADxVOP(OP_PADAV, OPf_MOD|OPf_REF, padix)));
+          break;
+
+        case '%':
+          ctx.bodyop = newBINOP(OP_AASSIGN, 0,
+            force_list_keeping_pushmark(newUNOP(OP_RV2AV, 0, newGVOP(OP_GV, 0, PL_defgv))),
+            force_list_keeping_pushmark(newPADxVOP(OP_PADHV, OPf_MOD|OPf_REF, padix)));
+          break;
+      }
+
+      ctx.retop = newLISTOP(OP_RETURN, 0,
+        newOP(OP_PUSHMARK, 0),
+        newPADxVOP(OP_PADSV, 0, PADIX_SELF));
+
+      break;
+    }
+
+    case ACCESSOR_LVALUE_MUTATOR:
+    {
+      assert(sigil == '$');
+
+      CvLVALUE_on(PL_compcv);
+
+      ctx.retop = newLISTOP(OP_RETURN, 0,
+        newOP(OP_PUSHMARK, 0),
+        newPADxVOP(OP_PADSV, 0, padix));
+
+      break;
+    }
+
+    case ACCESSOR_COMBINED:
+    {
+      assert(sigil == '$');
+
+      /* $field = shift if @_ */
+      ctx.bodyop = newLOGOP(OP_AND, 0,
+        /* scalar @_ */
+        op_contextualize(newUNOP(OP_RV2AV, 0, newGVOP(OP_GV, 0, PL_defgv)), G_SCALAR),
+        /* $field = shift */
+        newBINOP(OP_SASSIGN, 0,
+          newOP(OP_SHIFT, 0),
+          newPADxVOP(OP_PADSV, 0, padix)));
+
+      ctx.retop = newLISTOP(OP_RETURN, 0,
+        newOP(OP_PUSHMARK, 0),
+        newPADxVOP(OP_PADSV, 0, padix));
+
+      break;
+    }
+  }
 
   MOP_FIELD_RUN_HOOKS(fieldmeta, gen_accessor_ops, type, &ctx);
 
@@ -573,8 +742,6 @@ static void S_generate_field_accessor_method(pTHX_ FieldMeta *fieldmeta, SV *mna
   if(ctx.post_bodyops)
     ops = op_append_list(OP_LINESEQ, ops, ctx.post_bodyops);
 
-  if(!ctx.retop)
-    croak("Require ctx.retop");
   ops = op_append_list(OP_LINESEQ, ops, ctx.retop);
 
   SvREFCNT_inc(PL_compcv);
@@ -599,29 +766,10 @@ static void fieldhook_reader_seal(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void
   S_generate_field_accessor_method(aTHX_ fieldmeta, attrdata, ACCESSOR_READER);
 }
 
-static void fieldhook_gen_reader_ops(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
-{
-  if(type != ACCESSOR_READER)
-    return;
-
-  OPCODE optype = 0;
-
-  switch(SvPVX(fieldmeta->name)[0]) {
-    case '$': optype = OP_PADSV; break;
-    case '@': optype = OP_PADAV; break;
-    case '%': optype = OP_PADHV; break;
-  }
-
-  ctx->retop = newLISTOP(OP_RETURN, 0,
-    newOP(OP_PUSHMARK, 0),
-    newPADxVOP(optype, 0, ctx->padix));
-}
-
 static struct FieldHookFuncs fieldhooks_reader = {
-  .ver              = OBJECTPAD_ABIVERSION,
-  .apply            = &fieldhook_reader_apply,
-  .seal             = &fieldhook_reader_seal,
-  .gen_accessor_ops = &fieldhook_gen_reader_ops,
+  .ver   = OBJECTPAD_ABIVERSION,
+  .apply = &fieldhook_reader_apply,
+  .seal  = &fieldhook_reader_seal,
 };
 
 /* :writer */
@@ -637,41 +785,10 @@ static void fieldhook_writer_seal(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void
   S_generate_field_accessor_method(aTHX_ fieldmeta, attrdata, ACCESSOR_WRITER);
 }
 
-static void fieldhook_gen_writer_ops(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
-{
-  if(type != ACCESSOR_WRITER)
-    return;
-
-  switch(SvPVX(fieldmeta->name)[0]) {
-    case '$':
-      ctx->bodyop = newBINOP(OP_SASSIGN, 0,
-        newOP(OP_SHIFT, 0),
-        newPADxVOP(OP_PADSV, 0, ctx->padix));
-      break;
-
-    case '@':
-      ctx->bodyop = newBINOP(OP_AASSIGN, 0,
-        force_list_keeping_pushmark(newUNOP(OP_RV2AV, 0, newGVOP(OP_GV, 0, PL_defgv))),
-        force_list_keeping_pushmark(newPADxVOP(OP_PADAV, OPf_MOD|OPf_REF, ctx->padix)));
-      break;
-
-    case '%':
-      ctx->bodyop = newBINOP(OP_AASSIGN, 0,
-        force_list_keeping_pushmark(newUNOP(OP_RV2AV, 0, newGVOP(OP_GV, 0, PL_defgv))),
-        force_list_keeping_pushmark(newPADxVOP(OP_PADHV, OPf_MOD|OPf_REF, ctx->padix)));
-      break;
-  }
-
-  ctx->retop = newLISTOP(OP_RETURN, 0,
-    newOP(OP_PUSHMARK, 0),
-    newPADxVOP(OP_PADSV, 0, PADIX_SELF));
-}
-
 static struct FieldHookFuncs fieldhooks_writer = {
-  .ver              = OBJECTPAD_ABIVERSION,
-  .apply            = &fieldhook_writer_apply,
-  .seal             = &fieldhook_writer_seal,
-  .gen_accessor_ops = &fieldhook_gen_writer_ops,
+  .ver   = OBJECTPAD_ABIVERSION,
+  .apply = &fieldhook_writer_apply,
+  .seal  = &fieldhook_writer_seal,
 };
 
 /* :mutator */
@@ -691,23 +808,10 @@ static void fieldhook_mutator_seal(pTHX_ FieldMeta *fieldmeta, SV *attrdata, voi
   S_generate_field_accessor_method(aTHX_ fieldmeta, attrdata, ACCESSOR_LVALUE_MUTATOR);
 }
 
-static void fieldhook_gen_mutator_ops(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
-{
-  if(type != ACCESSOR_LVALUE_MUTATOR)
-    return;
-
-  CvLVALUE_on(PL_compcv);
-
-  ctx->retop = newLISTOP(OP_RETURN, 0,
-    newOP(OP_PUSHMARK, 0),
-    newPADxVOP(OP_PADSV, 0, ctx->padix));
-}
-
 static struct FieldHookFuncs fieldhooks_mutator = {
-  .ver              = OBJECTPAD_ABIVERSION,
-  .apply            = &fieldhook_mutator_apply,
-  .seal             = &fieldhook_mutator_seal,
-  .gen_accessor_ops = &fieldhook_gen_mutator_ops,
+  .ver   = OBJECTPAD_ABIVERSION,
+  .apply = &fieldhook_mutator_apply,
+  .seal  = &fieldhook_mutator_seal,
 };
 
 /* :accessor */
@@ -717,30 +821,29 @@ static void fieldhook_accessor_seal(pTHX_ FieldMeta *fieldmeta, SV *attrdata, vo
   S_generate_field_accessor_method(aTHX_ fieldmeta, attrdata, ACCESSOR_COMBINED);
 }
 
-static void fieldhook_gen_accessor_ops(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
+static struct FieldHookFuncs fieldhooks_accessor = {
+  .ver   = OBJECTPAD_ABIVERSION,
+  .apply = &fieldhook_mutator_apply, /* generate method name the same as :mutator */
+  .seal  = &fieldhook_accessor_seal,
+};
+
+/* :inheritable */
+
+static bool fieldhook_inheritble_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **attrdata_ptr, void *_funcdata)
 {
-  if(type != ACCESSOR_COMBINED)
-    return;
+  HV *hints = GvHV(PL_hintgv);
+  if(!hv_fetchs(hints, "Object::Pad/experimental(inherit_field)", 0))
+    Perl_ck_warner(aTHX_ packWARN(WARN_EXPERIMENTAL),
+      "inheriting fields is experimental and may be changed or removed without notice");
 
-  /* $field = shift if @_ */
-  ctx->bodyop = newLOGOP(OP_AND, 0,
-    /* scalar @_ */
-    op_contextualize(newUNOP(OP_RV2AV, 0, newGVOP(OP_GV, 0, PL_defgv)), G_SCALAR),
-    /* $field = shift */
-    newBINOP(OP_SASSIGN, 0,
-      newOP(OP_SHIFT, 0),
-      newPADxVOP(OP_PADSV, 0, ctx->padix)));
+  fieldmeta->is_inheritable = true;
 
-  ctx->retop = newLISTOP(OP_RETURN, 0,
-    newOP(OP_PUSHMARK, 0),
-    newPADxVOP(OP_PADSV, 0, ctx->padix));
+  return false;
 }
 
-static struct FieldHookFuncs fieldhooks_accessor = {
-  .ver              = OBJECTPAD_ABIVERSION,
-  .apply            = &fieldhook_mutator_apply, /* generate method name the same as :mutator */
-  .seal             = &fieldhook_accessor_seal,
-  .gen_accessor_ops = &fieldhook_gen_accessor_ops,
+static struct FieldHookFuncs fieldhooks_inheritable = {
+  .ver   = OBJECTPAD_ABIVERSION,
+  .apply = &fieldhook_inheritble_apply,
 };
 
 void ObjectPad_register_field_attribute(pTHX_ const char *name, const struct FieldHookFuncs *funcs, void *funcdata)
@@ -776,4 +879,7 @@ void ObjectPad__boot_fields(pTHX)
   register_field_attribute("writer",   &fieldhooks_writer,   NULL);
   register_field_attribute("mutator",  &fieldhooks_mutator,  NULL);
   register_field_attribute("accessor", &fieldhooks_accessor, NULL);
+
+  // TODO: temporary name
+  register_field_attribute("inheritable", &fieldhooks_inheritable, NULL);
 }
